/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user.keycloak;

import java.util.Optional;
import java.util.Spliterators;
import java.util.function.Supplier;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

import org.apache.commons.lang3.ObjectUtils;
import org.eclipse.microprofile.config.inject.ConfigProperty;
import org.keycloak.admin.client.resource.RealmResource;
import org.keycloak.admin.client.resource.UserResource;
import org.keycloak.representations.idm.UserRepresentation;

import de.ozgcloud.common.logging.OzgCloudLogging;
import de.ozgcloud.user.RemoteUserIterator;
import de.ozgcloud.user.User;
import de.ozgcloud.user.UserResourceMapper;
import de.ozgcloud.user.common.errorhandling.KeycloakUnavailableException;
import jakarta.enterprise.context.ApplicationScoped;
import jakarta.inject.Inject;
import jakarta.ws.rs.BadRequestException;
import jakarta.ws.rs.ClientErrorException;
import jakarta.ws.rs.NotFoundException;
import jakarta.ws.rs.ProcessingException;
import lombok.extern.log4j.Log4j2;

@ApplicationScoped
@OzgCloudLogging
@Log4j2
class KeycloakApiService {

	@Inject
	RealmResource realmResource;
	@Inject
	UserResourceMapper mapper;
	@Inject
	KeycloakApiProperties properties;

	@ConfigProperty(name = "keycloak.url")
	String keycloakUrl;

	public Stream<User> findAllUser() {
		return handlingKeycloakException(() -> getAllUserRepresentation().map(this::toUser));
	}

	Stream<UserRepresentation> getAllUserRepresentation() {
		var iterator = createRealmUserIterator();
		return StreamSupport.stream(Spliterators.spliteratorUnknownSize(iterator, 0), false);
	}

	Optional<User> findUserById(String id) {
		try {
			return Optional.of(realmResource.users().get(id)).map(mapper::toKopUser);
		} catch (NotFoundException e) {
			return Optional.empty();
		}
	}

	private RemoteUserIterator createRealmUserIterator() {
		return new RemoteUserIterator(realmResource);
	}

	private User toUser(UserRepresentation userRepresentation) {
		return mapper.toKopUser(realmResource.users().get(userRepresentation.getId()));
	}

	<T> T handlingKeycloakException(Supplier<T> runnable) {
		try {
			return runnable.get();
		} catch (ClientErrorException | ProcessingException | IllegalStateException e) {
			throw new KeycloakUnavailableException(properties.user(), properties.realm(), keycloakUrl, e);
		}
	}

	public void updateAttribute(String keycloakUserId, String attributeName, String attributeValue) {
		var userResource = realmResource.users().get(keycloakUserId);
		var userRepresentation = userResource.toRepresentation();
		var ozgCloudUserIdInKeycloak = userRepresentation.firstAttribute(attributeName);
		if (ObjectUtils.notEqual(ozgCloudUserIdInKeycloak, attributeValue)) {
			userRepresentation.singleAttribute(attributeName, attributeValue);
			tryUpdateUserResource(userResource, userRepresentation, attributeName);
		}
	}

	void tryUpdateUserResource(UserResource userResource, UserRepresentation userRepresentation, String attributeName) {
		try {
			userResource.update(userRepresentation);
		} catch (BadRequestException e) {
			LOG.warn("Could not update user attribute {} in Keycloak.", attributeName, e);
		}
	}
}