/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user.settings;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.UUID;

import jakarta.ws.rs.core.UriInfo;

import org.bson.types.ObjectId;
import org.eclipse.microprofile.jwt.JsonWebToken;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.user.User;
import de.ozgcloud.user.UserService;
import de.ozgcloud.user.UserTestFactory;
import de.ozgcloud.user.common.errorhandling.AccessForbiddenException;
import de.ozgcloud.user.common.errorhandling.FunctionalException;
import io.quarkus.hal.HalEntityWrapper;

class UserSettingsResourceTest {

	@Spy
	@InjectMocks
	private UserSettingsResource resource;

	@Mock
	private JsonWebToken jwt;

	@Mock
	private UserService userService;

	@Mock
	private UserSettingsService userSettingsService;

	@Mock
	private UriInfo uriInfo;

	@Mock
	private UserSettingsResourceAssembler resourceAssembler;

	private final String USER_MANAGER_URL = LoremIpsum.getInstance().getUrl();
	private final String USER_ID = UUID.randomUUID().toString();

	@BeforeEach
	void setUp() {
		resource.userManagerUrl = USER_MANAGER_URL;
	}

	@DisplayName("Get Usersettings")
	@Nested
	class TestGetUserSettings {

		private UserSettings userSettings = UserSettingsTestFactory.create();

		@BeforeEach
		void mockAccess() {
			doNothing().when(resource).checkUserAccess(anyString());
			when(userSettingsService.getByUserId(anyString())).thenReturn(userSettings);
		}

		@Test
		void shouldCheckAccess() {
			resource.getUserSettings(USER_ID);

			verify(resource).checkUserAccess(USER_ID);
		}

		@Test
		void shouldCallUserSettingsService() {
			resource.getUserSettings(USER_ID);

			verify(userSettingsService).getByUserId(USER_ID);
		}

		@Test
		void shouldCallResourceAssembler() {
			resource.getUserSettings(USER_ID);

			verify(resourceAssembler).toResource(userSettings, USER_ID, USER_MANAGER_URL);
		}
	}

	@DisplayName("Update Usersettings")
	@Nested
	class TestUpdateUserSettings {

		private final UserSettings userSettings = UserSettingsTestFactory.create();
		private final UserSettings updatedUserSettings = UserSettingsTestFactory.create();

		@DisplayName("with empty body")
		@Nested
		class TestOnEmptyBody {

			@BeforeEach
			void mockAccess() {
				doNothing().when(resource).checkUserAccess(anyString());
			}

			@Test
			void shouldThrowFunctionalExceptionOnMissingBody() {
				assertThatExceptionOfType(FunctionalException.class).isThrownBy(() -> resource.updateUserSettings(USER_ID, null))
						.withMessageStartingWith("Functional error: Request Body missing");
			}
		}

		@Nested
		class TestUserIdNotExist {
			@BeforeEach
			void mockAccess() {
				doNothing().when(resource).checkUserAccess(anyString());
			}

			@Test
			void shouldThrowFunctionalExceptionIfInvalidUserId() {
				var user = UserSettingsTestFactory.create();
				assertThatExceptionOfType(FunctionalException.class)
						.isThrownBy(() -> resource.updateUserSettings("wrong_id", user))
						.withMessageStartingWith("Functional error: Invalid user id");
			}

			@Test
			void shouldThrowFunctionalExceptionIfEmptyUserId() {
				var user = UserSettingsTestFactory.create();
				assertThatExceptionOfType(FunctionalException.class)
						.isThrownBy(() -> resource.updateUserSettings("", user))
						.withMessageStartingWith("Functional error: Invalid user id");
			}
		}

		@DisplayName("with filled body")
		@Nested
		class TestFilledBody {

			@BeforeEach
			void mockAccess() {
				doNothing().when(resource).checkUserAccess(anyString());
				when(userSettingsService.updateByUserId(anyString(), any(UserSettings.class))).thenReturn(Optional.of(updatedUserSettings));
			}

			@Test
			void shouldCheckAccess() {
				when(resourceAssembler.toResource(any(), anyString(), anyString())).thenReturn(new HalEntityWrapper(null));
				resource.updateUserSettings(USER_ID, userSettings);

				verify(resource).checkUserAccess(USER_ID);
			}

			@Test
			void shouldCallUserSettingsService() {
				when(resourceAssembler.toResource(any(), anyString(), anyString())).thenReturn(new HalEntityWrapper(null));
				resource.updateUserSettings(USER_ID, userSettings);

				verify(userSettingsService).updateByUserId(USER_ID, userSettings);
			}

			@Test
			void
			shouldCallResourceAssembler() {
				when(resourceAssembler.toResource(any(), anyString(), anyString())).thenReturn(new HalEntityWrapper(null));
				resource.updateUserSettings(USER_ID, userSettings);

				verify(resourceAssembler).toResource(updatedUserSettings, USER_ID, USER_MANAGER_URL);
			}
		}
	}

	@DisplayName("Update Usersettings")
	@Nested
	class TestPutUserSettings {

		private final UserSettings userSettings = UserSettingsTestFactory.create();
		private final UserSettings updatedUserSettings = UserSettingsTestFactory.create();

		@BeforeEach
		void mockAccess() {
			doNothing().when(resource).checkUserAccess(anyString());
		}

		@Test
		void shouldCheckAccess() {
			when(userSettingsService.updateByUserId(USER_ID, userSettings)).thenReturn(Optional.of(updatedUserSettings));
			when(resourceAssembler.toResource(any(), anyString(), anyString())).thenReturn(new HalEntityWrapper(null));

			resource.putUserSettings(USER_ID, userSettings);

			verify(resource).checkUserAccess(USER_ID);
		}

		@DisplayName("with empty body")
		@Nested
		class TestOnEmptyBody {

			@Test
			void shouldThrowFunctionalExceptionOnMissingBody() {
				assertThatExceptionOfType(FunctionalException.class).isThrownBy(() -> resource.putUserSettings(USER_ID, null))
						.withMessageStartingWith("Functional error: Request Body missing");
			}
		}

		@Nested
		class TestUserIdNotExist {

			@Test
			void shouldThrowFunctionalException() {
				when(userSettingsService.updateByUserId(USER_ID, userSettings)).thenReturn(Optional.empty());

				assertThatExceptionOfType(FunctionalException.class)
						.isThrownBy(() -> resource.putUserSettings(USER_ID, userSettings))
						.withMessageStartingWith("Functional error: Invalid user id");
			}
		}

		@Nested
		class TestFilledBody {

			@BeforeEach
			void setUp() {
				when(userSettingsService.updateByUserId(USER_ID, userSettings)).thenReturn(Optional.of(updatedUserSettings));
			}

			@Test
			void shouldCallUserSettingsService() {
				when(resourceAssembler.toResource(updatedUserSettings, USER_ID, USER_MANAGER_URL)).thenReturn(new HalEntityWrapper(null));

				resource.putUserSettings(USER_ID, userSettings);

				verify(userSettingsService).updateByUserId(USER_ID, userSettings);
			}

			@Test
			void
			shouldCallResourceAssembler() {
				when(resourceAssembler.toResource(updatedUserSettings, USER_ID, USER_MANAGER_URL)).thenReturn(new HalEntityWrapper(null));

				resource.putUserSettings(USER_ID, userSettings);

				verify(resourceAssembler).toResource(updatedUserSettings, USER_ID, USER_MANAGER_URL);
			}
		}
	}

	@DisplayName("Check User access")
	@Nested
	class TestCheckUserAccess {

		private final User user = UserTestFactory.createBuilder().id(new ObjectId()).build();

		@BeforeEach
		void init() {
			when(jwt.getSubject()).thenReturn(UserTestFactory.EXTERNAL_ID);
			when(userService.findByExternalId(anyString())).thenReturn(user);
		}

		@DisplayName("with existing user")
		@Nested
		class TestOnExistingUser {

			@BeforeEach
			void init() {
				when(userService.findByExternalId(anyString())).thenReturn(user);
			}

			@Test
			void shouldCallUserService() {
				resource.checkUserAccess(user.getId().toHexString());

				verify(userService).findByExternalId(anyString());
			}

			@Test
			void shouldThrowAccessForbiddenExceptionOnNoAccess() {
				assertThatExceptionOfType(AccessForbiddenException.class).isThrownBy(() -> resource.getUserSettings(USER_ID))
						.withMessageStartingWith("Functional error: Access to UserSettings is forbidden for user id");
			}
		}

		@DisplayName("with non existing user")
		@Nested
		class TestOnNonExistingUser {

			@Test
			void shouldCallUserService() {
				when(userService.findByExternalId(anyString())).thenReturn(user);

				resource.checkUserAccess(user.getId().toHexString());

				verify(userService).findByExternalId(UserTestFactory.EXTERNAL_ID);
			}

			@Test
			void shouldThrowAccessForbiddenExceptionOnMissingUser() {
				when(userService.findByExternalId(anyString())).thenReturn(null);

				assertThatExceptionOfType(AccessForbiddenException.class).isThrownBy(() -> resource.checkUserAccess(USER_ID))
						.withMessageStartingWith("Functional error: Access to UserSettings is forbidden for user id");
			}
		}
	}
}