/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.user.sync;

import static de.ozgcloud.user.User.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.time.Instant;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Stream;

import de.ozgcloud.user.TestDatabaseUtils;
import jakarta.inject.Inject;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import com.mongodb.client.MongoClient;
import com.mongodb.client.MongoDatabase;

import de.ozgcloud.user.User;
import de.ozgcloud.user.UserService;
import de.ozgcloud.user.UserTestFactory;
import de.ozgcloud.user.common.MongoDbTestProfile;
import de.ozgcloud.user.keycloak.KeycloakUserRemoteService;
import io.quarkus.test.InjectMock;
import io.quarkus.test.junit.QuarkusTest;
import io.quarkus.test.junit.TestProfile;
import io.quarkus.test.junit.mockito.InjectSpy;

@QuarkusTest
@TestProfile(MongoDbTestProfile.class)
class SyncServiceITCase {
	private static final String USER_KEY = "User";

	@InjectSpy
	SyncService service;

	@Inject
	MongoClient mongoClient;

	@Inject
	UserService userService;

	@InjectMock
	KeycloakUserRemoteService keycloakUserRemoteService;

	private MongoDatabase mongoDb;

	@BeforeEach
	void init() {
		when(keycloakUserRemoteService.getAllUsers()).thenReturn(Stream.of(UserTestFactory.create()));

		mongoDb = TestDatabaseUtils.getDatabase(mongoClient);
		mongoDb.drop();
	}

	@Test
	void shouldMarkNotSyncUsersAsDeleted() {
		prepareData();

		service.sync(Instant.now().toEpochMilli());

		prepareKeycloakUserWithoutRoles();

		service.sync(Instant.now().toEpochMilli());

		var syncedUser = mongoDb.getCollection(USER_KEY).find().first();
		assertThat(syncedUser.getBoolean(DELETED_FIELD)).isTrue();
	}

	private void prepareKeycloakUserWithoutRoles() {
		User user = UserTestFactory.create();
		when(keycloakUserRemoteService.getAllUsers()).thenReturn(Stream.of(user.toBuilder().roles(List.of()).build()));
	}

	@Test
	void shouldSync() {
		prepareData();

		service.sync(Instant.now().toEpochMilli());

		var syncedUsers = mongoDb.getCollection(USER_KEY).find().iterator();
		var val = syncedUsers.next().getLong(LAST_SYNC_TIMESTAMP_FIELD);
		assertThat(val).isGreaterThan(UserTestFactory.LAST_SYNC_TIMESTAMP);
	}

	private void prepareData() {
		User user1 = UserTestFactory.createBuilder().externalId("aaaaa").build();
		User user2 = UserTestFactory.createBuilder().externalId("bbbbb").build();
		persist(user1, user2);
		when(keycloakUserRemoteService.getAllUsers()).thenReturn(Stream.of(user1, user2));
	}

	private void persist(User... users) {
		Arrays.asList(users).forEach((user) -> {
			userService.save(user);
		});
	}
}
